#!/usr/bin/env python

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
IBM Z HMC OS Message Forwarder
"""

import argparse
import sys
import time
import logging
import logging.handlers

import urllib3
import zhmcclient

from ._version import __version__
from .forwarder_server import ForwarderServer
from . import utils  # for global variable VERBOSE_LEVEL
from .utils import DEFAULT_CONFIG_FILE, VALID_LOG_DESTINATIONS, \
    VALID_LOG_LEVELS, VALID_LOG_COMPONENTS, DEFAULT_LOG_LEVEL, \
    DEFAULT_LOG_COMP, DEFAULT_SYSLOG_FACILITY, VALID_SYSLOG_FACILITIES, \
    PRINT_ALWAYS, PRINT_V, RETRY_TIMEOUT_CONFIG, \
    ProperExit, ImproperExit, EarlyExit, \
    parse_yaml_file, logprint, setup_logging


def parse_args(args):
    """
    Parses the CLI arguments.
    """
    parser = argparse.ArgumentParser(
        description="IBM Z HMC OS Message Forwarder")
    parser.add_argument("-c", metavar="CONFIG_FILE",
                        default=DEFAULT_CONFIG_FILE,
                        help="path name of config file. "
                        "Use --help-config for details. "
                        "Default: {}".format(DEFAULT_CONFIG_FILE))
    parser.add_argument("--log", dest='log_dest', metavar="DEST", default=None,
                        help="enable logging and set a log destination "
                        "({dests}). Default: no logging".
                        format(dests=', '.join(VALID_LOG_DESTINATIONS)))
    parser.add_argument("--log-comp", dest='log_complevels', action='append',
                        metavar="COMP[=LEVEL]", default=None,
                        help="set a logging level ({levels}, default: "
                        "{def_level}) for a component ({comps}). May be "
                        "specified multiple times; options add to the default "
                        "of: {def_comp}".
                        format(levels=', '.join(VALID_LOG_LEVELS),
                               comps=', '.join(VALID_LOG_COMPONENTS),
                               def_level=DEFAULT_LOG_LEVEL,
                               def_comp=DEFAULT_LOG_COMP))
    parser.add_argument("--syslog-facility", metavar="TEXT",
                        default=DEFAULT_SYSLOG_FACILITY,
                        help="syslog facility ({slfs}) when logging to the "
                        "system log. Default: {def_slf}".
                        format(slfs=', '.join(VALID_SYSLOG_FACILITIES),
                               def_slf=DEFAULT_SYSLOG_FACILITY))
    parser.add_argument("--verbose", "-v", action='count', default=0,
                        help="increase the verbosity level (max: 2)")
    parser.add_argument("--version", action='store_true',
                        help="show versions of forwarder and zhmcclient "
                        "library and exit")
    parser.add_argument("--help-config", action='store_true',
                        help="show help for forwarder config file and exit")
    return parser.parse_args(args)


def print_version():
    """
    Print the version of this program and the zhmcclient library.
    """
    # pylint: disable=no-member
    print("zhmc_os_forwarder version: {}\n"
          "zhmcclient version: {}".
          format(__version__, zhmcclient.__version__))


def help_config():
    """
    Print help for the forwarder config file.
    """
    print("""
Help for forwarder config file

The forwarder config file is a YAML file that defines which HMC to talk to,
and the forwarding, i.e. which partition is forwarded to which syslog server.

The following example shows a complete forwarder config file. For more details,
see the documentation at https://zhmc-os-forwarder.readthedocs.io/.

---
hmc:
  hmc: 10.11.12.13
  userid: "myuser"
  password: "mypassword"
  verify_cert: false

forwarding:
  - syslogs:
     - server: 10.11.12.14
    cpcs:
      - cpc: MYCPC
        partitions:
          - partition: ".*"
""")


def main():
    """
    Main function for the script.
    """

    args = parse_args(sys.argv[1:])

    if args.version:
        print_version()
        sys.exit(0)

    if args.help_config:
        help_config()
        sys.exit(0)

    utils.VERBOSE_LEVEL = args.verbose

    urllib3.disable_warnings()

    forwarder_server = None

    try:
        setup_logging(args.log_dest, args.log_complevels, args.syslog_facility)

        logprint(logging.WARNING, None,
                 "---------------- "
                 "zhmc_os_forwarder started "
                 "----------------")

        logprint(logging.INFO, PRINT_ALWAYS,
                 f"zhmc_os_forwarder version: {__version__}")

        # pylint: disable=no-member
        logprint(logging.INFO, PRINT_ALWAYS,
                 f"zhmcclient version: {zhmcclient.__version__}")

        logprint(logging.INFO, PRINT_ALWAYS,
                 f"Verbosity level: {utils.VERBOSE_LEVEL}")

        config_filename = args.c
        logprint(logging.INFO, PRINT_V,
                 f"Parsing forwarder config file: {config_filename}")
        config_data = parse_yaml_file(
            config_filename, 'forwarder config file', 'config_schema.yaml')

        logprint(logging.INFO, PRINT_V,
                 "Timeout/retry configuration: "
                 "connect: {r.connect_timeout} sec / {r.connect_retries} "
                 "retries, read: {r.read_timeout} sec / {r.read_retries} "
                 "retries.".format(r=RETRY_TIMEOUT_CONFIG))

        forwarder_server = ForwarderServer(config_data, config_filename)
        try:
            forwarder_server.startup()
        except zhmcclient.Error as exc:
            new_exc = ImproperExit(
                f"{exc.__class__.__name__}: {exc}")
            new_exc.__cause__ = None  # pylint: disable=invalid-name
            raise new_exc

        logprint(logging.INFO, PRINT_V,
                 "Current number of subscriptions for OS message "
                 "notifications: {}".
                 format(forwarder_server.num_subscriptions))

        logprint(logging.INFO, PRINT_ALWAYS,
                 "Forwarder is up and running (Press Ctrl-C to shut down)")

        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                raise ProperExit

    except KeyboardInterrupt:
        logprint(logging.WARNING, PRINT_ALWAYS,
                 "Forwarder interrupted before server start")
        exit_rc(1)
    except EarlyExit as exc:
        logprint(logging.ERROR, PRINT_ALWAYS,
                 f"Error: {exc}")
        exit_rc(1)
    except ImproperExit as exc:
        logprint(logging.ERROR, PRINT_ALWAYS,
                 f"Error: {exc}")
        if forwarder_server:
            forwarder_server.shutdown()
        exit_rc(1)
    except ProperExit:
        logprint(logging.WARNING, PRINT_ALWAYS,
                 "Forwarder shutdown requested")
        if forwarder_server:
            forwarder_server.shutdown()
        exit_rc(0)


def exit_rc(rc):
    """Exit the script"""
    logprint(logging.WARNING, None,
             "---------------- "
             "zhmc_os_forwarder terminated "
             "----------------")
    sys.exit(rc)


if __name__ == "__main__":
    main()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions
"""

import sys
import os
import types
import platform
import time
import logging
from contextlib import contextmanager

import yaml
import jsonschema
import zhmcclient

#
# GLobal variables that will be set after command line parsing and will
# then be used throughout the project.
#

# Verbosity level from the command line
VERBOSE_LEVEL = 0

# Indicates that logging was enabled on the command line
LOGGING_ENABLED = False


#
# Logging
#

LOGGER_NAME = 'zhmcosforwarder'

# Logger names by log component
LOGGER_NAMES = {
    'forwarder': LOGGER_NAME,
    'hmc': zhmcclient.HMC_LOGGER_NAME,
    'jms': zhmcclient.JMS_LOGGER_NAME,
}
VALID_LOG_COMPONENTS = list(LOGGER_NAMES.keys()) + ['all']

# Log levels by their CLI names
LOG_LEVELS = {
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'off': logging.NOTSET,
}
VALID_LOG_LEVELS = list(LOG_LEVELS.keys())

# Defaults for --log-comp option.
DEFAULT_LOG_LEVEL = 'warning'
DEFAULT_LOG_COMP = 'all=warning'

# Values for printing messages dependent on verbosity level in command line
PRINT_ALWAYS = 0
PRINT_V = 1
PRINT_VV = 2

VALID_LOG_DESTINATIONS = ['stderr', 'syslog', 'FILE']

# Syslog facilities
VALID_SYSLOG_FACILITIES = [
    'user', 'local0', 'local1', 'local2', 'local3', 'local4', 'local5',
    'local6', 'local7'
]
DEFAULT_SYSLOG_FACILITY = 'user'

# Values to use for the 'address' parameter when creating a SysLogHandler.
# Key: Operating system type, as returned by platform.system(). For CygWin,
# the returned value is 'CYGWIN_NT-6.1', which is special-cased to 'CYGWIN_NT'.
# Value: Value for the 'address' parameter.
SYSLOG_ADDRESS = {
    'Linux': '/dev/log',
    'Darwin': '/var/run/syslog',  # macOS / OS-X
    'Windows': ('localhost', 514),
    'CYGWIN_NT': '/dev/log',  # Requires syslog-ng pkg
    'other': ('localhost', 514),  # used if no key matches
}


#
# Defaults for other command line options
#

DEFAULT_CONFIG_FILE = '/etc/zhmc-os-forwarder/config.yaml'


#
# Retry
#

# Sleep time in seconds when retrying HMC connections
RETRY_SLEEP_TIME = 10

# Retry / timeout configuration for zhmcclient (used at the socket level)
RETRY_TIMEOUT_CONFIG = zhmcclient.RetryTimeoutConfig(
    connect_timeout=10,
    connect_retries=2,
    read_timeout=300,
    read_retries=2,
    max_redirects=zhmcclient.DEFAULT_MAX_REDIRECTS,
    operation_timeout=zhmcclient.DEFAULT_OPERATION_TIMEOUT,
    status_timeout=zhmcclient.DEFAULT_STATUS_TIMEOUT,
    name_uri_cache_timetolive=zhmcclient.DEFAULT_NAME_URI_CACHE_TIMETOLIVE,
)


#
# Exceptions
#

class ConfigFileError(Exception):
    """
    An error was found in the config file.
    """
    pass


class ConnectionError(Exception):
    # pylint: disable=redefined-builtin
    """
    Connection error with the HMC.
    """
    pass


class AuthError(Exception):
    """
    Authentication error with the HMC.
    """
    pass


class OtherError(Exception):
    """
    Other error with the HMC.
    """
    pass


class ProperExit(Exception):
    """
    Terminating while the server was running.
    """
    pass


class ImproperExit(Exception):
    """
    Terminating because something went wrong.
    """
    pass


class EarlyExit(Exception):
    """
    Terminating before the server was started.
    """
    pass


@contextmanager
def zhmc_exceptions(session, config_filename):
    # pylint: disable=invalid-name
    """
    Context manager that handles zhmcclient exceptions by raising the
    appropriate forwarder exceptions.

    Example::

        with zhmc_exceptions(session, config_filename):
            client = zhmcclient.Client(session)
            version_info = client.version_info()
    """
    try:
        yield
    except zhmcclient.ConnectionError as exc:
        new_exc = ConnectionError(
            "Connection error using IP address {} defined in forwarder config "
            "file {}: {}".format(session.host, config_filename, exc))
        new_exc.__cause__ = None
        raise new_exc  # ConnectionError
    except zhmcclient.ClientAuthError as exc:
        new_exc = AuthError(
            "Client authentication error for the HMC at {h} using "
            "userid '{u}' defined in forwarder config file {f}: {m}".
            format(h=session.host, u=session.userid, f=config_filename,
                   m=exc))
        new_exc.__cause__ = None
        raise new_exc  # AuthError
    except zhmcclient.ServerAuthError as exc:
        http_exc = exc.details  # zhmcclient.HTTPError
        new_exc = AuthError(
            "Authentication error returned from the HMC at {h} using "
            "userid '{u}' defined in forwarder config file {f}: {m} "
            "(HMC operation {hm} {hu}, HTTP status {hs}.{hr})".
            format(h=session.host, u=session.userid, f=config_filename,
                   m=exc, hm=http_exc.request_method, hu=http_exc.request_uri,
                   hs=http_exc.http_status, hr=http_exc.reason))
        new_exc.__cause__ = None
        raise new_exc  # AuthError
    except OSError as exc:
        new_exc = OtherError(str(exc))
        new_exc.__cause__ = None
        raise new_exc  # OtherError
    except zhmcclient.Error as exc:
        new_exc = OtherError(
            f"Error returned from HMC at {session.host}: {exc}")
        new_exc.__cause__ = None
        raise new_exc  # OtherError


def validate_option(option_name, option_value, allowed_values):
    """
    Validate the option value against the allowed option values
    and return the value, if it passes. raises EarlyExit otherwise.

    Raises:
      EarlyExit: Invalid command line usage.
    """
    if option_value not in allowed_values:
        raise EarlyExit(
            "Invalid value {val} for {opt} option. Allowed are: {allowed}".
            format(opt=option_name, val=option_value,
                   allowed=', '.join(allowed_values)))
    return option_value


def parse_yaml_file(yamlfile, name, schemafilename=None):
    """
    Returns the parsed content of a YAML file as a Python object.
    Optionally validates against a specified JSON schema file in YAML format.

    Raises:
        ImproperExit
    """

    try:
        with open(yamlfile, encoding='utf-8') as fp:
            yaml_obj = yaml.safe_load(fp)
    except FileNotFoundError as exc:
        new_exc = ImproperExit(
            "Cannot find {} {}: {}".
            format(name, yamlfile, exc))
        new_exc.__cause__ = None  # pylint: disable=invalid-name
        raise new_exc
    except PermissionError as exc:
        new_exc = ImproperExit(
            "Permission error reading {} {}: {}".
            format(name, yamlfile, exc))
        new_exc.__cause__ = None  # pylint: disable=invalid-name
        raise new_exc
    except yaml.YAMLError as exc:
        new_exc = ImproperExit(
            "YAML error reading {} {}: {}".
            format(name, yamlfile, exc))
        new_exc.__cause__ = None  # pylint: disable=invalid-name
        raise new_exc

    if schemafilename:

        schemafile = os.path.join(
            os.path.dirname(__file__), 'schemas', schemafilename)
        try:
            with open(schemafile, encoding='utf-8') as fp:
                schema = yaml.safe_load(fp)
        except FileNotFoundError as exc:
            new_exc = ImproperExit(
                "Internal error: Cannot find schema file {}: {}".
                format(schemafile, exc))
            new_exc.__cause__ = None  # pylint: disable=invalid-name
            raise new_exc
        except PermissionError as exc:
            new_exc = ImproperExit(
                "Internal error: Permission error reading schema file {}: {}".
                format(schemafile, exc))
            new_exc.__cause__ = None  # pylint: disable=invalid-name
            raise new_exc
        except yaml.YAMLError as exc:
            new_exc = ImproperExit(
                "Internal error: YAML error reading schema file {}: {}".
                format(schemafile, exc))
            new_exc.__cause__ = None  # pylint: disable=invalid-name
            raise new_exc

        try:
            jsonschema.validate(yaml_obj, schema)
        except jsonschema.exceptions.SchemaError as exc:
            new_exc = ImproperExit(
                "Internal error: Invalid JSON schema file {}: {}".
                format(schemafile, exc))
            new_exc.__cause__ = None
            raise new_exc
        except jsonschema.exceptions.ValidationError as exc:
            element_str = json_path_str(exc.absolute_path)
            new_exc = ImproperExit(
                "Validation of {} {} failed on {}: {}".
                format(name, yamlfile, element_str, exc.message))
            new_exc.__cause__ = None
            raise new_exc

    return yaml_obj


def json_path_str(path_list):
    """
    Return a string with the path list in JSON path notation, except that
    the root element is not '$' but verbally expressed.
    """
    if not path_list:
        return "root elements"

    path_str = ""
    for p in path_list:
        if isinstance(p, int):
            path_str += f"[{p}]"
        else:
            path_str += f".{p}"
    if path_str.startswith('.'):
        path_str = path_str[1:]
    return f"element '{path_str}'"


def get_hmc_info(session):
    """
    Return the result of the 'Query API Version' operation. This includes
    the HMC version, HMC name and other data. For details, see the operation's
    result description in the HMC WS API book.

    Raises: zhmccclient exceptions
    """
    client = zhmcclient.Client(session)
    hmc_info = client.query_api_version()
    return hmc_info


def logprint(log_level, print_level, message):
    """
    Log a message at the specified log level, and print the message at
    the specified verbosity level

    Parameters:
        log_level (int): Python logging level at which the message should be
          logged (logging.DEBUG, etc.), or None for no logging.
        print_level (int): Verbosity level at which the message should be
          printed (1, 2), or None for no printing.
        message (string): The message.
    """
    if print_level is not None and VERBOSE_LEVEL >= print_level:
        print(message)
    if log_level is not None and LOGGING_ENABLED:
        logger = logging.getLogger(LOGGER_NAME)
        # Note: This method never raises an exception. Errors during logging
        # are handled by calling handler.handleError().
        logger.log(log_level, message)


def setup_logging(log_dest, log_complevels, syslog_facility):
    """
    Set up Python logging as specified in the command line.

    Raises:
        EarlyExit
    """
    global LOGGING_ENABLED  # pylint: disable=global-statement

    if log_dest is None:
        logprint(None, PRINT_V, "Logging is disabled")
        handler = None
        dest_str = None
    elif log_dest == 'stderr':
        dest_str = "the Standard Error stream"
        logprint(None, PRINT_V, f"Logging to {dest_str}")
        handler = logging.StreamHandler(stream=sys.stderr)
    elif log_dest == 'syslog':
        system = platform.system()
        if system.startswith('CYGWIN_NT'):
            # Value is 'CYGWIN_NT-6.1'; strip off trailing version:
            system = 'CYGWIN_NT'
        try:
            address = SYSLOG_ADDRESS[system]
        except KeyError:
            address = SYSLOG_ADDRESS['other']
        dest_str = ("the System Log at address {a!r} with syslog facility "
                    "{slf!r}".format(a=address, slf=syslog_facility))
        logprint(None, PRINT_V, f"Logging to {dest_str}")
        try:
            facility = logging.handlers.SysLogHandler.facility_names[
                syslog_facility]
        except KeyError:
            valid_slfs = ', '.join(
                logging.handlers.SysLogHandler.facility_names.keys())
            raise EarlyExit(
                "This system ({sys}) does not support syslog facility {slf}. "
                "Supported are: {slfs}.".
                format(sys=system, slf=syslog_facility, slfs=valid_slfs))
        # The following does not raise any exception if the syslog address
        # cannot be opened. In that case, the first attempt to log something
        # will fail.
        handler = logging.handlers.SysLogHandler(
            address=address, facility=facility)
    else:
        dest_str = f"file {log_dest}"
        logprint(None, PRINT_V, f"Logging to {dest_str}")
        try:
            handler = logging.FileHandler(log_dest)
        except OSError as exc:
            raise EarlyExit(
                "Cannot log to file {fn}: {exc}: {msg}".
                format(fn=log_dest, exc=exc.__class__.__name__, msg=exc))

    if not handler and log_complevels:
        raise EarlyExit(
            "--log-comp option cannot be used when logging is disabled; "
            "use --log option to enable logging.")

    if handler:

        def handleError(self, record):
            """
            Replacement for built-in method on logging.Handler class.

            This is needed because the SysLogHandler class does not raise
            an exception when creating the handler object, but only when
            logging something to it.
            """
            _, exc, _ = sys.exc_info()
            f_record = self.format(record)
            print("Error: Logging to {d} failed with: {exc}: {msg}. Formatted "
                  "log record: {r!r}".
                  format(d=dest_str, exc=exc.__class__.__name__, msg=exc,
                         r=f_record),
                  file=sys.stderr)
            sys.exit(1)

        handler.handleError = types.MethodType(handleError, handler)

        logger_level_dict = {}  # key: logger_name, value: level
        if not log_complevels:
            log_complevels = [DEFAULT_LOG_COMP]
        for complevel in log_complevels:
            if '=' in complevel:
                comp, level = complevel.split('=', 2)
            else:
                comp = complevel
                level = DEFAULT_LOG_LEVEL
            if level not in LOG_LEVELS:
                raise EarlyExit(
                    "Invalid log level {level!r} in --log-comp option. "
                    "Allowed are: {allowed}".
                    format(level=level, allowed=', '.join(VALID_LOG_LEVELS)))
            if comp == 'all':
                for logger_name in LOGGER_NAMES.values():
                    logger_level_dict[logger_name] = level
            else:
                try:
                    logger_name = LOGGER_NAMES[comp]
                except KeyError:
                    raise EarlyExit(
                        "Invalid component {comp!r} in --log-comp option. "
                        "Allowed are: {allowed}".
                        format(comp=comp,
                               allowed=', '.join(VALID_LOG_COMPONENTS)))
                logger_level_dict[logger_name] = level

        complevels = ', '.join(
            [f"{name}={level}"
             for name, level in logger_level_dict.items()])
        logprint(None, PRINT_V,
                 "Logging components: {complevels}".
                 format(complevels=complevels))

        if isinstance(handler, logging.handlers.SysLogHandler):
            # Most syslog implementations fail when the message is longer
            # than a limit. We use a hard coded limit for now:
            # * 2048 is the typical maximum length of a syslog message,
            #   including its headers
            # * 41 is the max length of the syslog message parts before MESSAGE
            # * 47 is the max length of the Python format string before message
            # Example syslog message:
            #   <34>1 2003-10-11T22:14:15.003Z localhost MESSAGE
            # where MESSAGE is the formatted Python log message.
            max_msg = f'.{2048 - 41 - 47}'
        else:
            max_msg = ''
        fs = ('%(asctime)s %(levelname)s %(name)s: %(message){m}s'.
              format(m=max_msg))

        # Set the formatter to always log times in UTC. Since the %z
        # formatting string does not get adjusted for that, set the timezone
        # offset always to '+0000'.
        dfs = '%Y-%m-%d %H:%M:%S+0000'
        logging.Formatter.converter = time.gmtime  # log times in UTC

        handler.setFormatter(logging.Formatter(fmt=fs, datefmt=dfs))
        for logger_name in LOGGER_NAMES.values():
            logger = logging.getLogger(logger_name)
            if logger_name in logger_level_dict:
                level = logger_level_dict[logger_name]
                level_int = LOG_LEVELS[level]
                if level_int != logging.NOTSET:
                    logger.addHandler(handler)
                logger.setLevel(level_int)
            else:
                logger.setLevel(logging.NOTSET)

        LOGGING_ENABLED = True



#!/usr/bin/env python3

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the parse_args() function.
"""

from zhmc_os_forwarder.zhmc_os_forwarder import parse_args
from zhmc_os_forwarder.utils import DEFAULT_CONFIG_FILE, \
    DEFAULT_SYSLOG_FACILITY


def test_parse_args_full():
    """
    Test with full set of options.
    """
    cmdline = [
        "-c", "foo.yml",
        "-v", "-v",
        "--log", "syslog",
        "--log-comp", "all=debug",
        "--syslog-facility", "local0",
    ]
    args = parse_args(cmdline)
    assert args.c == "foo.yml"
    assert args.log_dest == "syslog"
    assert "all=debug" in args.log_complevels
    assert args.syslog_facility == "local0"


def test_parse_args_min():
    """
    Test with minimal set of options.
    """
    args = parse_args([])
    assert args.c == DEFAULT_CONFIG_FILE
    assert args.log_dest is None
    assert args.log_complevels is None  # default set later
    assert args.syslog_facility == DEFAULT_SYSLOG_FACILITY



#!/usr/bin/env python

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A class for storing forwarded LPARs and their syslog servers
"""

from .forwarder_config import ForwarderConfig


# pylint: disable=too-few-public-methods
class ForwardedLparInfo:
    """
    Info for a single forwarded LPAR
    """

    def __init__(self, lpar, syslogs=None, topic=None):
        self.lpar = lpar
        if not syslogs:
            syslogs = []
        self.syslogs = syslogs
        self.topic = topic


class ForwardedLpars:
    """
    A data structure to maintain forwarded LPARs and their syslog servers,
    based on the forwarder config.
    """

    def __init__(self, session, config_data, config_filename):
        """
        Parameters:
          session (zhmcclient.Session): Session with the HMC.
          config_data (dict): Content of forwarder config file.
          config_filename (string): Path name of forwarder config file.
        """
        self.session = session
        self.config_data = config_data
        self.config_filename = config_filename

        # Forwarder config for fast lookup
        self.config = ForwarderConfig(config_data, config_filename)

        # Representation of forwarded LPARs
        # - key: LPAR URI
        # - value: ForwardedLparInfo
        self.forwarded_lpar_infos = {}

    def __str__(self):
        return ("{s.__class__.__name__}("
                "config_filename={s.config_filename!r}"
                ")".format(s=self))

    def __repr__(self):
        return ("{s.__class__.__name__}("
                "config_filename={s.config_filename!r}, "
                "config={s.config!r}, "
                "forwarded_lpar_infos={s.forwarded_lpar_infos!r}"
                ")".format(s=self))

    def add_if_matching(self, lpar):
        """
        Add an LPAR to be forwarded if it matches a forwarding definition
        in the forwarder config.

        If the LPAR is already being forwarded, its syslog servers are changed
        to the syslog servers from the forwarder definition.

        Parameters:
          lpar (zhmcclient.Partition/Lpar or string): The LPAR, as a zhmcclient
            resource object or as a URI string.

        Returns:
            bool: Indicates whether the LPAR was added.
        """
        syslogs = self.config.get_syslogs(lpar)
        if syslogs:
            if lpar.uri not in self.forwarded_lpar_infos:
                self.forwarded_lpar_infos[lpar.uri] = ForwardedLparInfo(lpar)
            self.forwarded_lpar_infos[lpar.uri].syslogs = syslogs
            return True
        return False

    def remove(self, lpar):
        """
        Remove an LPAR from being forwarded.

        If the LPAR is not currently being forwarded, Nothing is done.

        Parameters:
          lpar (zhmcclient.Partition/Lpar or string): The LPAR, as a zhmcclient
            resource object or as a URI string.
        """
        if lpar.uri in self.forwarded_lpar_infos:
            del self.forwarded_lpar_infos[lpar.uri]

    def is_forwarding(self, lpar):
        """
        Return whether the LPAR is currently being forwarded.

        Parameters:
          lpar (zhmcclient.Partition/Lpar or string): The LPAR, as a zhmcclient
            resource object or as a URI string.

        Returns:
          bool: Indicates whether the LPAR is currently being forwarded.
        """
        return lpar.uri in self.forwarded_lpar_infos

    def get_syslogs(self, lpar):
        """
        Get the syslogs from the forwarder config for a forwarded LPAR.

        If the LPAR is not currently forwarded, returns None.

        Parameters:
          lpar (zhmcclient.Partition/Lpar or string): The LPAR, as a zhmcclient
            resource object or as a URI string.

        Returns:
          list of ConfigSyslogInfo: The syslogs for the LPAR, or None.
        """
        try:
            lpar_info = self.forwarded_lpar_infos[lpar.uri]
        except KeyError:
            return None
        return lpar_info.syslogs



#!/usr/bin/env python3

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the parse_yaml_file() function.
"""

import sys
import os
import stat
import time
import hashlib

import pytest

from zhmc_os_forwarder.utils import parse_yaml_file, ImproperExit


def test_parse_yaml_file_simple():
    """
    Tests if a simple forwarder config file is correctly parsed.
    """
    # Get a SHA256 of Unixtime to create a filename that does not exist
    filename = str(hashlib.sha256(str(time.time()).encode("utf-8")).
                   hexdigest())
    with open(filename, "w+", encoding='utf-8') as testfile:
        testfile.write("""
hmc:
  host: 10.11.12.13
  userid: "myuser"
  password: "mypassword"
  verify_cert: false

forwarding:
  - syslogs:
     - server: 10.11.12.14
    cpcs:
      - cpc: MYCPC
        partitions:
          - partition: ".*"
""")

    expected_result = {
        "hmc": {
            "host": "10.11.12.13",
            "userid": "myuser",
            "password": "mypassword",
            "verify_cert": False,
        },
        "forwarding": [
            {
                "syslogs": [
                    {
                        "server": "10.11.12.14",
                    },
                ],
                "cpcs": [
                    {
                        "cpc": "MYCPC",
                        "partitions": [
                            {
                                "partition": ".*",
                            },
                        ],
                    },
                ],
            },
        ],
    }
    result = parse_yaml_file(filename, 'test file')
    assert result == expected_result

    os.remove(filename)


def test_parse_yaml_file_permission():
    """
    Tests if permission denied is correctly handled.
    """

    if sys.platform == 'win32':
        pytest.skip("Test not supported on Windows")

    filename = str(hashlib.sha256(str(time.time()).encode("utf-8")).
                   hexdigest())
    with open(filename, "w+", encoding='utf-8'):
        pass
    # Make it unreadable (mode 000)
    os.chmod(filename, not stat.S_IRWXU)
    with pytest.raises(ImproperExit):
        parse_yaml_file(filename, 'test file')
    os.remove(filename)


def test_parse_yaml_file_notfound():
    """
    Tests if file not found is correctly handled.
    """
    filename = str(hashlib.sha256(str(time.time()).encode("utf-8")).
                   hexdigest())
    with pytest.raises(ImproperExit):
        parse_yaml_file(filename, 'test file')



#!/usr/bin/env python

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A class for running the forwarder server
"""

import os
import logging
import socket
from threading import Thread, Event

import zhmcclient

from .forwarded_lpars import ForwardedLpars
from .utils import logprint, PRINT_ALWAYS, PRINT_V, PRINT_VV, \
    RETRY_TIMEOUT_CONFIG


class ForwarderServer:
    """
    A forwarder server.
    """

    def __init__(self, config_data, config_filename):
        """
        Parameters:
          config_data (dict): Content of forwarder config file.
          config_filename (string): Path name of forwarder config file.
        """
        self.config_data = config_data
        self.config_filename = config_filename

        self.thread = Thread(target=self.run)  # forwarder thread
        self.thread_started = False
        self.stop_event = Event()  # Set event to stop forwarder thread

        self.session = None  # zhmcclient.Session with the HMC

        self.all_cpcs = None  # List of all managed CPCs as zhmcclient.Cpc
        self.all_lpars = None  # List of all partitions/LPARs as zhmcclient obj

        self.forwarded_lpars = None  # ForwardedLpars object

        self.receiver = None  # NotificationReceiver
        self.num_subscriptions = None

    def startup(self):
        """
        Set up the forwarder server and start the forwarder thread.
        """

        hmc_data = self.config_data['hmc']
        # hmc data structure in config file:
        #   hmc:
        #     hmc: 10.11.12.13
        #     userid: "myuser"
        #     password: "mypassword"
        #     verify_cert: false

        verify_cert = hmc_data.get('verify_cert', True)
        if isinstance(verify_cert, str):
            if not os.path.isabs(verify_cert):
                verify_cert = os.path.join(
                    os.path.dirname(self.config_filename), verify_cert)

        logprint(logging.INFO, PRINT_ALWAYS,
                 "Opening session with HMC {h} "
                 "(user: {u}, certificate validation: {c})".
                 format(h=hmc_data['host'], u=hmc_data['userid'],
                        c=verify_cert))

        self.session = zhmcclient.Session(
            hmc_data['host'],
            hmc_data['userid'],
            hmc_data['password'],
            verify_cert=verify_cert,
            retry_timeout_config=RETRY_TIMEOUT_CONFIG)

        client = zhmcclient.Client(self.session)

        logprint(logging.INFO, PRINT_V,
                 "Gathering information about CPCs and LPARs to forward")
        self.all_cpcs = client.cpcs.list()
        self.all_lpars = []
        for cpc in self.all_cpcs:
            dpm = cpc.prop('dpm-enabled')
            if dpm:
                self.all_lpars.extend(cpc.partitions.list())
            else:
                self.all_lpars.extend(cpc.lpars.list())

        self.forwarded_lpars = ForwardedLpars(
            self.session, self.config_data, self.config_filename)

        for lpar in self.all_lpars:
            cpc = lpar.manager.parent
            added = self.forwarded_lpars.add_if_matching(lpar)
            if added:
                logprint(logging.INFO, PRINT_V,
                         "LPAR {p!r} on CPC {c!r} will be forwarded".
                         format(p=lpar.name, c=cpc.name))

        self.receiver = zhmcclient.NotificationReceiver(
            [],  # self.session.object_topic to get notifications to ignore
            hmc_data['host'],
            hmc_data['userid'],
            hmc_data['password'])

        self.num_subscriptions = 0
        logger_id = 0  # ID number used in Python logger name
        for lpar_info in self.forwarded_lpars.forwarded_lpar_infos.values():
            lpar = lpar_info.lpar
            cpc = lpar.manager.parent

            logprint(logging.INFO, PRINT_VV,
                     "Opening OS message channel for LPAR {p!r} on CPC {c!r}".
                     format(p=lpar.name, c=cpc.name))
            try:
                os_topic = lpar.open_os_message_channel(
                    include_refresh_messages=True)
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 409 and exc.reason == 331:
                    # OS message channel is already open for this session,
                    # reuse its notification topic.
                    topic_dicts = self.session.get_notification_topics()
                    os_topic = None
                    for topic_dict in topic_dicts:
                        if topic_dict['topic-type'] != \
                                'os-message-notification':
                            continue
                        obj_uri = topic_dict['object-uri']
                        if lpar.uri == obj_uri:
                            os_topic = topic_dict['topic-name']
                            logprint(logging.INFO, PRINT_VV,
                                     "Using existing OS message notification "
                                     "topic {t!r} for LPAR {p!r} on CPC {c!r}".
                                     format(t=os_topic, p=lpar.name,
                                            c=cpc.name))
                            break
                    if os_topic is None:
                        raise RuntimeError(
                            "An OS message notification topic for LPAR {p!r} "
                            "on CPC {c!r} supposedly exists, but cannot be "
                            "found in the existing topics for this session: "
                            "{t}".
                            format(p=lpar.name, c=cpc.name, t=topic_dicts))
                elif exc.http_status == 409 and exc.reason == 332:
                    # The OS does not support OS messages.
                    logprint(logging.WARNING, PRINT_ALWAYS,
                             "Warning: The OS in LPAR {p!r} on CPC {c!r} does "
                             "not support OS messages - ignoring the LPAR".
                             format(p=lpar.name, c=cpc.name))
                    os_topic = None
                else:
                    raise

            if os_topic:
                logprint(logging.INFO, PRINT_VV,
                         "Subscribing for OS message notifications for LPAR "
                         "{p!r} on CPC {c!r} (topic: {t})".
                         format(p=lpar.name, c=cpc.name, t=os_topic))
                self.receiver.subscribe(os_topic)
                lpar_info.topic = os_topic
                self.num_subscriptions += 1

            # Prepare sending to syslogs by creating Python loggers
            for syslog in self.forwarded_lpars.get_syslogs(lpar):
                try:
                    logger = self._create_logger(syslog, logger_id)
                except ConnectionError as exc:
                    logprint(logging.WARNING, PRINT_ALWAYS,
                             f"Warning: Skipping syslog server: {exc}")
                    continue
                logger_id += 1
                syslog.logger = logger

        self._start()
        self.thread_started = True

    @staticmethod
    def _create_logger(syslog, logger_id):
        facility_code = logging.handlers.SysLogHandler.facility_names[
            syslog.facility]
        if syslog.port_type == 'tcp':
            # Newer syslog protocols, e.g. rsyslog
            socktype = socket.SOCK_STREAM
        else:
            assert syslog.port_type == 'udp'
            # Older syslog protocols, e.g. BSD
            socktype = socket.SOCK_DGRAM
        try:
            handler = logging.handlers.SysLogHandler(
                (syslog.host, syslog.port), facility_code,
                socktype=socktype)
        # pylint: disable=broad-exception-caught
        except Exception as exc:
            raise ConnectionError(
                "Cannot create log handler for syslog server at "
                "{host}, port {port}/{port_type}: {msg}".
                format(host=syslog.host, port=syslog.port,
                       port_type=syslog.port_type, msg=str(exc)))
        handler.setFormatter(logging.Formatter('%(message)s'))
        logger_name = f'zhmcosfwd_syslog_{logger_id}'
        logger = logging.getLogger(logger_name)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        return logger

    def shutdown(self):
        """
        Stop the forwarder thread and clean up the forwarder server.
        """

        if self.forwarded_lpars:
            for lpar_info in self.forwarded_lpars.forwarded_lpar_infos.values():
                lpar = lpar_info.lpar
                cpc = lpar.manager.parent

                if lpar_info.topic:
                    logprint(logging.INFO, PRINT_VV,
                             "Unsubscribing OS message channel for LPAR {p!r} "
                             "on CPC {c!r} (topic: {t})".
                             format(p=lpar.name, c=cpc.name, t=lpar_info.topic))
                    try:
                        self.receiver.unsubscribe(lpar_info.topic)
                    except zhmcclient.Error as exc:
                        logprint(logging.ERROR, PRINT_ALWAYS,
                                 "Error unsubscribing OS message channel for "
                                 "LPAR {p!r} on CPC {c!r} (topic: {t}): {m}".
                                 format(p=lpar.name, c=cpc.name,
                                        t=lpar_info.topic, m=exc))

        if self.receiver:
            try:
                logprint(logging.INFO, PRINT_ALWAYS,
                         "Closing notification receiver")
                self.receiver.close()
            except zhmcclient.Error as exc:
                logprint(logging.ERROR, PRINT_ALWAYS,
                         "Error closing notification receiver: {m}".
                         format(m=exc))

        if self.thread_started:
            try:
                logprint(logging.INFO, PRINT_ALWAYS,
                         "Stopping forwarder thread")
                self._stop()
            # pylint: disable=broad-exception-caught
            except Exception as exc:
                logprint(logging.ERROR, PRINT_ALWAYS,
                         "Error stopping forwarder thread: {m}".
                         format(m=exc))
            self.thread_started = False

        # logprint(logging.INFO, PRINT_ALWAYS,
        #          "Cleaning up partition notifications on HMC")
        # for lpar_tuple in self.forwarded_lpars.values():
        #     lpar = lpar_tuple[0]
        #     try:
        #         lpar.disable_auto_update()
        #     except zhmcclient.HTTPError as exc:
        #         if exc.http_status == 403:
        #             # The session does not exist anymore
        #             pass

        if self.session:
            logprint(logging.INFO, PRINT_ALWAYS,
                     "Closing session with HMC")
            try:
                self.session.logoff()
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 403:
                    # The session does not exist anymore
                    pass
                else:
                    logprint(logging.ERROR, PRINT_ALWAYS,
                             "Error closing session with HMC: {m}".
                             format(m=exc))
            self.session = None

    def _start(self):
        """
        Start the forwarder thread.
        """
        self.stop_event.clear()
        self.thread.start()

    def _stop(self):
        """
        Stop the forwarder thread.
        """
        self.stop_event.set()
        self.thread.join()

    def run(self):
        """
        The method running as the forwarder server thread.
        """
        logprint(logging.INFO, PRINT_V,
                 "Entering forwarder thread")
        while True:

            if self.stop_event.is_set():
                break

            try:
                # pylint: disable=unused-variable
                for headers, message in self.receiver.notifications():
                    self.handle_notification(headers, message)

            except zhmcclient.NotificationJMSError as exc:
                logprint(logging.ERROR, PRINT_ALWAYS,
                         "Error receiving notifications {}: {}".
                         format(exc.__class__.__name__, exc))
                logprint(logging.ERROR, PRINT_ALWAYS,
                         "Receiving notifications again")

        logprint(logging.INFO, PRINT_V,
                 "Leaving forwarder thread")

    def handle_notification(self, headers, message):
        """
        Handle a received notification.
        """
        noti_type = headers['notification-type']
        if noti_type == 'os-message':
            for msg_info in message['os-messages']:
                lpar_uri = headers['object-uri']
                lpar_infos = self.forwarded_lpars.forwarded_lpar_infos
                lpar_info = lpar_infos[lpar_uri]
                lpar = lpar_info.lpar
                seq_no = msg_info['sequence-number']
                msg_txt = msg_info['message-text'].strip('\n')
                self.send_to_syslogs(lpar, seq_no, msg_txt)
        else:
            dest = headers['destination']
            sub_id = headers['subscription']
            obj_class = headers['class']
            obj_name = headers['name']
            logprint(logging.WARNING, PRINT_ALWAYS,
                     "Warning: Ignoring {nt!r} notification for {c} {n!r} "
                     "(subscription: {s}, destination: {d})".
                     format(nt=noti_type, c=obj_class, n=obj_name, s=sub_id,
                            d=dest))

    def send_to_syslogs(self, lpar, seq_no, msg_txt):
        """
        Send a single OS message to the configured syslogs for its LPAR.
        """
        cpc = lpar.manager.parent
        for syslog in self.forwarded_lpars.get_syslogs(lpar):
            if syslog.logger:
                syslog_txt = ('{c} {p} {s}: {m}'.
                              format(c=cpc.name, p=lpar.name, s=seq_no,
                                     m=msg_txt))
                try:
                    syslog.logger.info(syslog_txt)
                # pylint: disable=broad-exception-caught
                except Exception as exc:
                    logprint(logging.WARNING, PRINT_ALWAYS,
                             "Warning: Cannot send seq_no {s} from LPAR {p!r} "
                             "on CPC {c!r} to syslog host {h}: {m}".
                             format(s=seq_no, p=lpar.name, c=cpc.name,
                                    h=syslog.host, m=exc))
                    continue



#!/usr/bin/env python

# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A class for storing forwarded LPARs and their syslog servers
"""

import re

from collections import namedtuple

# Default syslog properties, if not specified in forwarder config
DEFAULT_SYSLOG_PORT = 514
DEFAULT_SYSLOG_PORT_TYPE = 'tcp'
DEFAULT_SYSLOG_FACILITY = 'user'

# Info for a single CPC pattern in the forwarder config
ConfigCpcInfo = namedtuple(
    'ConfigCpcInfo',
    [
        'cpc_pattern',          # string: Compiled pattern for CPC name
        'lpar_infos',           # List of ConfigLparInfo items
    ]
)


# Info for a single LPAR pattern in the forwarder config
# In context of a CPC pattern.
ConfigLparInfo = namedtuple(
    'ConfigLparInfo',
    [
        'lpar_pattern',         # string: Compiled pattern for LPAR name
        'syslogs',              # list of ConfigSyslogInfo: Syslogs for the LPAR
    ]
)


# pylint: disable=too-few-public-methods
class ConfigSyslogInfo:
    """
    Info for a single syslog in the forwarder config
    """

    def __init__(self, host, port, port_type, facility):
        self.host = host  # string: Syslog IP address or hostname
        self.port = port  # int: Syslog port number
        self.port_type = port_type  # int: Syslog port type ('tcp', 'udp')
        self.facility = facility  # string: Syslog facility (e.g. 'user')
        self.logger = None  # logging.Logger: Python logger for syslog


class ForwarderConfig:
    """
    A data structure to keep the forwarder config in an optimized way.
    """

    def __init__(self, config_data, config_filename):
        """
        Parameters:
          config_data (dict): Content of forwarder config file.
          config_filename (string): Path name of forwarder config file.
        """
        self.config_data = config_data
        self.config_filename = config_filename

        # Data for storing the config

        # Representation of forwarder config for fast lookup
        # - items: namedtuple ConfigCpcInfo
        self.config_cpc_infos = []

        forwarding = self.config_data['forwarding']
        # forwarding data structure in config file:
        #   forwarding:
        #     - syslogs:
        #        - server: 10.11.12.14
        #       cpcs:
        #         - cpc: CPC.*
        #           partitions:
        #             - partition: "dal1-.*"

        for fwd_item in forwarding:
            syslogs = []
            for sl_item in fwd_item['syslogs']:
                sl_host = sl_item['host']
                sl_port = sl_item.get('port', DEFAULT_SYSLOG_PORT)
                sl_port_type = sl_item.get('port_type',
                                           DEFAULT_SYSLOG_PORT_TYPE)
                sl_facility = sl_item.get('facility', DEFAULT_SYSLOG_FACILITY)
                syslog_info = ConfigSyslogInfo(
                    sl_host, sl_port, sl_port_type, sl_facility)
                syslogs.append(syslog_info)
            for cpc_item in fwd_item['cpcs']:
                cpc_pattern = re.compile('^{}$'.format(cpc_item['cpc']))
                cpc_info = ConfigCpcInfo(cpc_pattern, [])
                for lpar_item in cpc_item['partitions']:
                    lpar_pattern = re.compile(
                        '^{}$'.format(lpar_item['partition']))
                    lpar_info = ConfigLparInfo(lpar_pattern, syslogs)
                    cpc_info.lpar_infos.append(lpar_info)
                self.config_cpc_infos.append(cpc_info)

    def __str__(self):
        return ("{s.__class__.__name__}("
                "config_filename={s.config_filename!r}"
                ")".format(s=self))

    def __repr__(self):
        return ("{s.__class__.__name__}("
                "config_filename={s.config_filename!r}, "
                "config_cpc_infos={s.config_cpc_infos!r}"
                ")".format(s=self))

    def get_syslogs(self, lpar):
        """
        Get the syslogs for an LPAR if it matches the forwarder config.

        If it does not match the forwarder config, None is returned.

        Parameters:
          lpar (zhmcclient.Partition/Lpar): The LPAR, as a zhmcclient
            resource object.

        Returns:
          list of ConfigSyslogInfo: List of syslogs if matching, or None
          otherwise.
        """
        cpc = lpar.manager.parent
        for cpc_info in self.config_cpc_infos:
            if cpc_info.cpc_pattern.match(cpc.name):
                for lpar_info in cpc_info.lpar_infos:
                    if lpar_info.lpar_pattern.match(lpar.name):
                        return lpar_info.syslogs
        return None



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Setup script for the zhmc-os-forwarder project.
"""

import os
import setuptools


def get_version(version_file):
    """
    Execute the specified version file and return the value of the __version__
    global variable that is set in the version file.

    Note: Make sure the version file does not depend on any packages in the
    requirements list of this package (otherwise it cannot be executed in
    a fresh Python environment).
    """
    with open(version_file, encoding='utf-8') as fp:
        version_source = fp.read()
    _globals = {}
    exec(version_source, _globals)  # pylint: disable=exec-used
    return _globals['__version__']


def get_requirements(requirements_file):
    """
    Parse the specified requirements file and return a list of its non-empty,
    non-comment lines. The returned lines are without any trailing newline
    characters.
    """
    with open(requirements_file, encoding='utf-8') as fp:
        lines = fp.readlines()
    reqs = []
    for line in lines:
        line = line.strip('\n')
        if not line.startswith('#') and line != '':
            reqs.append(line)
    return reqs


def read_file(a_file):
    """
    Read the specified file and return its content as one string.
    """
    with open(a_file, encoding='utf-8') as fp:
        content = fp.read()
    return content


# pylint: disable=invalid-name
requirements = get_requirements('requirements.txt')
install_requires = requirements
package_version = get_version(
    os.path.join('zhmc_os_forwarder', '_version.py'))

# Docs on setup():
# * https://docs.python.org/3.8/distutils/apiref.html?
#   highlight=setup#distutils.core.setup
# * https://setuptools.readthedocs.io/en/latest/setuptools.html#
#   new-and-changed-setup-keywords
# Explanations for the behavior of package_data, include_package_data, and
# MANIFEST files:
# * https://setuptools.readthedocs.io/en/latest/setuptools.html#
#   including-data-files
# * https://stackoverflow.com/a/11848281/1424462
# * https://stackoverflow.com/a/14159430/1424462
setuptools.setup(
    name='zhmc_os_forwarder',
    version=package_version,
    packages=[
        'zhmc_os_forwarder'
    ],
    package_data={
        'zhmc_os_forwarder': ['schemas/*.yaml'],
    },
    entry_points={
        'console_scripts': [
            'zhmc_os_forwarder = '
            'zhmc_os_forwarder.zhmc_os_forwarder:main',
        ],
    },
    # include_package_data=True,  # Includes MANIFEST.in files into sdist (only)
    install_requires=install_requires,
    # dependency_links is deprecated, and pip since 18.1 supports PEP-508 URLs.

    description='IBM Z HMC OS Message Forwarder',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    license='Apache License, Version 2.0',
    author='Andreas Maier',
    author_email='maiera@de.ibm.com',
    maintainer='Andreas Maier, Kathir Velusamy',
    maintainer_email='maiera@de.ibm.com, kathir.velu@in.ibm.com',
    url='https://github.com/zhmcclient/zhmc-os-forwarder',
    project_urls={
        'Bug Tracker':
        'https://github.com/zhmcclient/zhmc-os-forwarder/issues',
        'Documentation':
        'https://zhmc-os-forwarder.readthedocs.io/en/stable/',
        'Change Log':
        'https://zhmc-os-forwarder.readthedocs.io/en/stable/'
        'changes.html',
    },

    options={'bdist_wheel': {'universal': True}},
    zip_safe=True,  # This package can safely be installed from a zip file
    platforms='any',

    # Keep these Python versions in sync with:
    # - Section "Supported environments" in docs/intro.rst
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ]
)
